#!/usr/bin/perl -w

#############################################################################
#
# This script provides weekly overview of email stored in the junkmail
# folder and allows for unjunking. When an email is being unjunked the
# baysian filter in SpamAssassin is trained as ham.
#
# This script has been developed
# by Jesper Knudsen at http://sme.swerts-knudsen.dk
#
# Revision History:
#
# August 24, 2008:      Initial version
#############################################################################

# internal modules (part of core perl distribution)
use Getopt::Long;
use Pod::Usage;
use POSIX qw/strftime floor/;
use Time::Local;
use Date::Manip;
use strict;
use Cwd "realpath";
use File::Basename;
use English;
use CGI qw(:standard);
use File::Copy;
use File::Spec;
use Data::Dumper;
use IO::Socket::INET;
use Carp;
use Getopt::Long;


use esmith::db;
use esmith::ConfigDB;
use esmith::AccountsDB;

my %optctl = ();
GetOptions (\%optctl, "duration=s","log","central","crontab","logfile=s");

if (not defined($optctl{"duration"})) {
  die "\nUsage: $0 -duration=<day|hour>\n";
};

my $log = $optctl{"log"} ? 1 : 0;
my $duration = $optctl{"duration"};
my $log_file = defined($optctl{"logfile"}) ? $optctl{"logfile"} : '/var/log/dansguardian/access.log';

# Temp. log location
my $log_location = '/var/log/dansguardian/';

my $release_version;

my $master_host_name = 'central.swerts-knudsen.dk';
my $master_host_port = '1111';
my $statsclient = 'enabled';
my $central = $optctl{"central"} ? 1 : 0;
my $crontab = $optctl{"crontab"} ? 1 : 0;

get_SME_version();
read_log();


# ---------------------------------

sub get_SME_version {

  my $dbh = esmith::ConfigDB->open() || die "Unable to open configuration dbase.";

  my %sa_conf = $dbh->get('sysconfig')->props;

  while (my ($parameter,$value) = each(%sa_conf)) {
    if ($parameter eq 'ReleaseVersion') {
      $release_version = $value;
    }
  }
}

sub _date {

  my $time = shift;

  $time = time() if (not defined($time));

  my @weekday = ("Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat");
  my @month = ("Jan", "Feb", "Mar", "Apr", "May", "Jun",
               "Jul", "Aug", "Sep", "Oct", "Nov", "Dec");

  my @date_info = localtime($time);
  my $date = sprintf("%s %d %s %4d, %02d:%02d:%02d",
                     $weekday[$date_info[6]],
                     $date_info[3], $month[$date_info[4]], $date_info[5] + 1900,
                     $date_info[2], $date_info[1], $date_info[0]);

  return $date;
}


sub read_log {

  my $YEAR = (localtime(time))[5]; # this is years since 1900
  my $start;
  my $end;

  my $seen = 0;
  my $seen_size = 0;
  my $scanned = 0;
  my $infected = 0;
  my $urlmod = 0;
  my $contentmod = 0;

  my $exception = 0;

  my $banned = 0;
  my $denied = 0;

  my $scanned_size = 0;
  my $average_score = 0;
  my $virus_size = 0;
  my $denied_size = 0;

  my $domain_list;
  my $domain_denied_list;
  my $virus_list;
  my $category_list;
  
  if (uc($duration) eq 'DAY') {
    ($start, $end) = parse_arg("yesterday", "");
  } else {
    ($start, $end) = parse_arg("1 hour ago", "");
  }
  
  
  if (open (READ, "< $log_file")) {
    
    foreach my $line (<READ>) {
      
      $line =~ s/[ \t\n]*$//;

      my @keys = split (/\t/,$line);
#      my @keys = split (/ /,$line);

#      printf("Line: %s \n",$line);
#      printf("Size: %s \n",scalar(@keys));

      next if (scalar(@keys) == 1);

#      printf("KEY0=%s\n",$keys[0]);

      # 2009.4.13 22:47:04 - 192.168.212.244 http://www.torrentflux.com *SCANNED*  GET 4729 -35  1 200 text/html   -
      
#      my ($year,$month,$day) = $keys[0] =~ m/^2([^.]+).([^.]+).([^ ]+)/;
      
#      my ($hour,$min,$sec) = $keys[1] =~ m/^([^:]+):([^:]+):(.*)/;


      my ($year,$month,$day,$hour,$min,$sec) = $keys[0] =~ m/^2([^.]+).([^.]+).([^ ]+) ([^:]+):([^:]+):(.*)/;

#      printf("%s-%s-%s - %s:%s:%s\n",$year,$month,$day,$hour,$min,$sec);

#      my ($rest);

#      my ($year,$month,$day,$hour,$min,$sec,$rest) = $line =~ m/^2([^.]+).([^.]+).([^ ]+) ([^:]+):([^:]+):([^ ]+) (.*)/;
 
      next if (not defined($sec));

      $year += 2000;

#      if (defined($year)) {
#       printf("%s-%s-%s - %s:%s:%s\n",$year,$month,$day,$hour,$min,$sec);
#      }

      # Convert to absolute time
      my $abstime = timelocal($sec, $min, $hour, $day, $month-1, $year);
      my $abshour = floor ($abstime / 3600); # Hours since the epoch
     
#      if (not defined($year)) {
#	printf("%s-%s-%s - %s:%s:%s\n",$year+2000,$month,$day,$hour,$min,$sec);
#      }

      #If date specified, only process lines matching date
      next if ($abstime < $start);
      # We can assume that logs are chronological
 #     last if ($abstime > $end);
      
 #     printf("MATCH: %s-%s-%s - %s:%s:%s\n",$year,$month,$day,$hour,$min,$sec);

      # 2009.4.13 22:47:04 - 192.168.212.244 http://www.torrentflux.com *SCANNED*  GET 4729 -35  1 200 text/html   -

 #     printf("IP=%s\n",$keys[2]);
#      printf("%s\n",$rest);
 
      my $url = $keys[3];
      last if (not defined($url));	

      # If we recognize the URL and date its an understood log format
      $seen++;
      my ($text,$rest1,$rest2) = $line =~ m/(.*)\t(GET|POST|HEAD|PUT|DELETE|TRACE|CONNECT)\t(.*)/;
      if (defined($rest2)) {
	my ($size,$weight,$rest3) = $rest2 =~ m/^([^\t]+)\t([^\t]+)\t(.*)/;	
	if (defined($size)) {
#	  printf("Size = %s",$size);
	  $seen_size += $size;
	}
      }
      
      $url =~ s!^https?://(?:www\.)?!!i;
      $url =~ s!/.*!!;
      $url =~ s/[\?\#\:].*//;

      # Check for IP URLs
      if ($url =~ m/([0-9]+).([0-9]+).([0-9]+).([0-9]+)/) {
#	printf("IP URL=%s\n",$url);
      } else {

	# Get rid of subdomains (f.ex "toolbarqueries" in "toolbarqueries.google.com")
	while ((() = $url =~ m/[.]/g) > 1) {
#	printf("More than one \".\" - %s\n",$url);
	  
	  # Let the .co.xx (ex. .co.uk) and .com.xx (ex. .com.br) domains have 3 dots....
	  last if (() = $url =~ m/[.]/g) == 2 and ($url =~ m/\.co\./ or $url =~ m/\.com\./);
	  my ($lead,$new) = $url =~ m/([^.]+).(.*)/;
#       printf("Now one \".\" - %s\n",$new);
	  $url = $new;
	}
      }

#      printf("URL=%s\n",$url);

      $domain_list->{$url}++;

      my $result = $keys[4];

#      printf("Res=%s",$result);
      if ($result =~ m/^\*SCANNED\*/) {

        my ($virus) = $line =~ m/\*SCANNED\* (.*)/;
        my $action;

#2009.4.14 21:11:48      -       192.168.212.244 http://www.sex.com/site/        *SCANNED* *DENIED* Weighted phrase limit of 250 : 398 ((-news, article)+(18 u.s.c., 2257)+(18, u.s.c, 2257,  compl)+( 2257 , statement)+(usc, 2257, statement)+ sex +adult movie+-advice+-movie review+sex story+sex toys+sex-shop+sexy)        GET     11778   398     Pornography, Pornography (German)       1       403     -  

        if ($virus =~ m/^\*DENIED\*/) {
          $virus =~ s/^\*DENIED\* //;
          $action = 'Denied';
	  $denied++;
        } elsif ($virus =~ m/^\*EXCEPTED\*/) {
          $virus =~ s/^\*EXCEPTED\* //;
          $action = 'Excepted';
	  $scanned++;
        } else {
	  $scanned++;
	}

	if (defined($action)) {
	  my ($text,$rest1,$rest2) = $virus =~ m/(.*)\t(GET|POST|HEAD|PUT|DELETE|TRACE|CONNECT)\t(.*)/;
	  my ($text1) = $text =~ m/([^\(]+)\((.*)/;
#	  printf("%s due to =\"%s\"\n",$action,$text1);
	}

	# Get size and score
	my ($text,$rest1,$rest2) = $line =~ m/(.*)\t(GET|POST|HEAD|PUT|DELETE|TRACE|CONNECT)\t(.*)/;
#	printf("TXT=%s,RST1=%s,RST2=%s\n",$text,$rest1,$rest2);
#	printf("RST2=%s\n",$rest2);
	# Its not GET|POST|HEAD but maybe OPTIONS|PROPFIND
	if (not defined($rest2)) {
#	  printf("LINE=%s\n",$line);
	  next;
	}
	my ($size,$weight,$rest3) = $rest2 =~ m/^([^\t]+)\t([^\t]+)\t(.*)/;	

#	printf("Size=%s, Score=%s\n",$size,$weight);

	if (defined($action)) {
#	  $rest3 =~ s/[\t]/#/;
#	  my ($category) = $rest3 =~ m/([^\t]+)/;
	  # Log format is hard to guess - either there is a group of categories and I only want the first
	  # or its on and I need to cut after before TAB.
	  my $category;
	  if ($rest3 =~ m/\,/) {
	    ($category) = $rest3 =~ m/([^,]+)/;
	  } else {
	    ($category) = $rest3 =~ m/([^\t]+)/;
	  }
	  $category_list->{$category}++;
#	  printf("Scanned and %s with score %s in Category \"%s\"\n",$action,$weight,$category);
	}

	if (defined($action) and $action eq 'Denied') {
	  $denied_size += $size;
	  $domain_denied_list->{$url}++;
	}

	$scanned_size += $size;
	$average_score += $weight;

      } elsif ($result =~m/^\*INFECTED\*/) {
        $infected++;

#2009.4.14 20:15:39	-	192.168.212.244	http://securite-informatique.info/virus/eicar/download/eicar.zip	*INFECTED* *DENIED* Virus or bad content detected. Eicar-Test-Signature	GET	271	0	Content scanning	1	403	application/zip			-

        my ($virus) = $line =~ m/\*INFECTED\* (.*)/;
	my $action;

	if ($virus =~ m/^\*DENIED\*/) {
	  $virus =~ s/^\*DENIED\* //;
	  $action = 'Denied';
	} elsif ($virus =~ m/^\*EXCEPTED\*/) {
	  $virus =~ s/^\*EXCEPTED\* //;
	  $action = 'Excepted';
	}

        my ($text,$rest1,$rest2) = $virus =~ m/(.*)\t(GET|POST|HEAD|PUT|DELETE|TRACE|CONNECT)\t(.*)/;
#        printf("%s due to =\"%s\"\n",$action,$text);

	# get rid of unwanted text - Need to fix language variants...
	$text =~ s/Virus or bad content detected. //;

	# Dutch
	$text =~ s/Virus of onbetamelijke inhoud ontdekt. //;


	$virus_list->{$text}++;

	my ($size,$weight,$rest3) = $rest2 =~ m/^([^\t]+)\t([^\t]+)\t(.*)/;	

	my ($category) = $rest3 =~ m/([^\t]+)/;

#	printf("Infected and %s with \"%s\" in category \"%s\"\n",$action,$text,$category);
	$category_list->{$category}++;

#	printf("Size=%s, Score=%s\n",$size,$weight);

#	$scanned++;
	$virus_size += $size;
#        $scanned_size += $size;
        $average_score += $weight;

#	my ($size,$weight,$category,$group,$http,$mime) = $rest2 =~ m/() (.*)/;	

      } elsif ($result =~m/^\*URLMOD\*/) {
        $urlmod++;
      } elsif ($result =~m/^\*CONTENTMOD\*/) {
        $contentmod++;
      } elsif ($result =~m/^\*EXCEPTION\*/) {
        $exception++;
	$scanned++;

        ($text,$rest1,$rest2) = $line =~ m/(.*)\t(GET|POST|HEAD|PUT|DELETE|TRACE|CONNECT)\t(.*)/;
	if (defined($rest2)) {
	  my ($size,$weight,$rest3) = $rest2 =~ m/^([^\t]+)\t([^\t]+)\t(.*)/;

	  if (defined($size)) {
	    $scanned_size += $size;	
	  }
	}

      } elsif ($result =~m/^\*DENIED\*/) {

#2009.4.14 16:15:14	-	192.168.212.244	http://sunmicro-4.vo.llnwd.net/u/ESD7/JSCDL/jdk/6u13-b90/jre/jre-6u13-windows-i586-p-iftw.exe	*DENIED* Banned extension: .exe	HEAD	0	0	Banned extension	1	403	application/download			-
	
        $denied++;

        my ($virus) = $line =~ m/\*DENIED\* (.*)/;
	my $action = 'Denied';

        my ($text1,$rest1,$rest2) = $virus =~ m/(.*)\t(GET|POST|HEAD|PUT|DELETE|TRACE|CONNECT)\t(.*)/;
#        printf("Denied due to =\"%s\" - %s\n",$text1,$action);

	my ($text2,$rest3,$rest4) = $line =~ m/(.*)\t(GET|POST|HEAD|PUT|DELETE|TRACE|CONNECT)\t(.*)/;

	my ($size,$weight,$rest5) = $rest4 =~ m/^([^\t]+)\t([^\t]+)\t(.*)/;	
	
#	printf("Size=%s, Score=%s\n",$size,$weight);
	
#       $rest1 =~ s/[\t]/#/;
#       printf("Rest=%s\n",$rest1);

	
	my ($category) = $text1 =~ m/^([^:]+)/;
#	printf("Denied and %s with score %s in Category \"%s\"\n %s\n",$action,$weight,$category,$line);
	
	$category_list->{$category}++;

	$domain_denied_list->{$url}++;

	$denied_size += $size;

	$average_score += $weight;


      } else {
	if ($result ne '' and not $crontab) {
	  printf("Different Action=%s\n",$result);
	  printf("LINE: %s\n",$line);
	}
      }

     
    }
    close(READ);
    
  } else {
    printf("Cannot open log file (%s) for reading!",$log_file);
    return;
  }
  
  if (not $crontab) {
    printf("---Overall result ----------------\n");
    printf("Start time    = %s\n",_date($start));
    printf("End time      = %s\n",_date());
    printf("Seen URLs     = %s (%s)\n",$seen,utils_convert_bytes_to_optimal_unit($seen_size));
    printf("Scanned URLs  = %s (%s)\n",$scanned,utils_convert_bytes_to_optimal_unit($scanned_size));
    printf("Infected      = %s (%s)\n",$infected,utils_convert_bytes_to_optimal_unit($virus_size));
    printf("URL mod       = %s\n",$urlmod);
    printf("Content mod   = %s\n",$contentmod);
    printf("Denied        = %s (%s)\n",$denied,utils_convert_bytes_to_optimal_unit($denied_size));
    
    if ($scanned > 0) {
      printf("--- Statistics ------------\n");
      printf("Average Score = %.2f\n",$average_score/$scanned);
      printf("Problems      = %.3f %%\n",($infected+$denied)/$scanned);

      printf("--- TOP 10 domains ------------\n");
      sort_domains($domain_list);

      if ($domain_denied_list) {
        printf("--- TOP 10 Blocked Domains -----\n");
        sort_domains($domain_denied_list);
      }

      if ($virus_list) {
	printf("--- TOP 10 Blocked Virus/Malware-------\n");
	sort_domains($virus_list);
      }
      if ($category_list) {
        printf("--- TOP 10 Blocked Categories-------\n");
        sort_domains($category_list);
      }

    }
  }

 # Now store the data for web statistics

  if ($log) {
    my $LogFile = sprintf ("%s/dansguardianstats.%s",$log_location,$duration);
    if (open (WRITE, "+> $LogFile")) {
      
#    printf WRITE ("START=%s|",$start);
#    printf WRITE ("END=%s|",$end);
      printf WRITE ("SMEVERSION=%s|",$release_version);
      printf WRITE ("SEEN=%s|",$seen);
      printf WRITE ("SEENSIZE=%s|",$seen_size);
      printf WRITE ("SCANNED=%s|",$scanned);
      printf WRITE ("SCANNEDSIZE=%s|",$scanned_size);
      printf WRITE ("INFECTED=%s|",$infected);
      printf WRITE ("INFECTEDSIZE=%s|",$virus_size);
      printf WRITE ("DENIED=%s|",$denied);
      printf WRITE ("DENIEDSIZE=%s|",$denied_size);
      printf WRITE ("URLMOD=%s|",$urlmod);
      printf WRITE ("CONTENTMOD=%s|",$contentmod);
      
      my @sorted_list = sort {$domain_list->{$b} <=> $domain_list->{$a}} keys %$domain_list;
      
      my $count = 1;
      foreach my $domain (@sorted_list) {
	printf WRITE ("DOMAIN=%s|",$domain);
	printf WRITE ("DCOUNT=%s|",$domain_list->{$domain});
	last if ($count++ >= 10);
      }

      $count = 1;
      @sorted_list = sort {$domain_denied_list->{$b} <=> $domain_denied_list->{$a}} keys %$domain_denied_list;
      foreach my $domain (@sorted_list) {
        printf WRITE ("DENIEDDOMAIN=%s|",$domain);
        printf WRITE ("DENIEDCOUNT=%s|",$domain_denied_list->{$domain});
        last if ($count++ >= 10);
      }


      $count = 1;
      @sorted_list = sort {$virus_list->{$b} <=> $virus_list->{$a}} keys %$virus_list;
      foreach my $virus (@sorted_list) {
        printf WRITE ("VIRUS=%s|",$virus);
        printf WRITE ("VCOUNT=%s|",$virus_list->{$virus});
        last if ($count++ >= 10);
      }

      $count = 1;
      @sorted_list = sort {$category_list->{$b} <=> $category_list->{$a}} keys %$category_list;
      foreach my $category (@sorted_list) {
        printf WRITE ("CATEGORY=%s|",$category);
        printf WRITE ("CCOUNT=%s|",$category_list->{$category});
        last if ($count++ >= 10);
      }
      printf WRITE ("\n");

      close(WRITE);
    }
    
    # Now send the statistics to the Statistice Server
    if ($central and uc($duration) eq 'HOUR' and uc($statsclient) eq 'ENABLED') {
      ConnectAndSend(CreateStatisticsMsg('web'));
    }

  }
}



sub sort_domains {

  my $domain_list = shift;

  my @sorted_list = sort {$domain_list->{$b} <=> $domain_list->{$a}} keys %$domain_list;

  my $count = 1;
  foreach my $domain (@sorted_list) {
    printf("%s (%s times)\n",$domain,$domain_list->{$domain});
    last if ($count++ >= 10);
  }
}

sub utils_convert_bytes_to_optimal_unit{
  my($bytes) = @_;

  return '' if ($bytes eq '');

  my($size);
  $size = $bytes . ' Bytes' if ($bytes < 1024);
  $size = sprintf("%.2f", ($bytes/1024)) . ' KB' if ($bytes >= 1024 && $bytes < 1048576);
  $size = sprintf("%.2f", ($bytes/1048576)) . ' MB' if ($bytes >= 1048576 && $bytes < 1073741824);
  $size = sprintf("%.2f", ($bytes/1073741824)) . ' GB' if ($bytes >= 1073741824 && $bytes < 1099511627776);
  $size = sprintf("%.2f", ($bytes/1099511627776)) . ' TB' if ($bytes >= 1099511627776);

  return $size;
}

sub parse_arg {
  my $startdate = shift;
  my $enddate = shift;

  my $secsinday = 86400;
  my $time = 0;

  my $start = UnixDate($startdate,"%s");
  my $end = UnixDate($enddate, "%s");

  if(!$start && !$end) {
    $end = time;
    $start = $end - $secsinday;
    return ($start, $end);
  }

  if(!$start) {
    $start = $end - $secsinday;
    return ($start, $end);
  }

  if(!$end) {
    $end = $start + $secsinday;
    return ($start, $end);
  }
  if($start > $end) {
    return ($end, $start);
  }

  return ($start, $end);

}


sub CreateStatisticsMsg {

  my $stattype = shift;

  # General
  my $encryption = 'NO';
  my $compression = 'NO';
  my $version = 1;


  # Prepare Header
  my $msg = sprintf("COMMAND=LOGGING\nTYPE=%s\nCOMPRESSION=%s\nENCRYPTION=%s\nVERSION=%s\n",uc($stattype),$compression,$encryption,$version);

  $msg .= 'DATA=';

#  printf("MSGHDR=%s\n",$msg);

  my $LogFile = sprintf ("%s/dansguardianstats.%s",$log_location,$duration);

  if (open (LOG, "< $LogFile")) {
    foreach my $line (<LOG>) {
 #     printf("TXT=%s\n",$line);
      $msg .= $line;
    }
    close(LOG);
  }

#  printf("MSG=%s",$msg);

  return $msg;
}

sub ConnectAndSend() {

  my $command =  shift;

  my $return_value = undef;
  my $retry_attempts = 3;

  my $socket = job_start_client($master_host_name,
                                $master_host_port,
                                120
                                );


  if (not defined($socket)) {
    return $return_value;
  }

  TRY_AGAIN:
#  printf("Initiating Contact with statistics server \"%s\" (%s)\n",$master_host_name,$retry_attempts);

  my $message = sprintf ("ACCOUNT=%s\|\n", GetMacAddress());

#  printf("Tx Msg = %s",$message);
  job_send_data($socket,$message);

  my $msg = <$socket>;

  if (not defined($msg)) {
    if ($retry_attempts-- > 0) {
      # Wait a little while before trying again if system seems ill....
      sleep(int(rand(10)));
#      debug(5,"Need a retry - %s retries left,",$retry_attempts);
      goto TRY_AGAIN;
    } else {
      job_stop_client($socket);
#      printf("Timed out connecting to master");
      return $return_value;
    }
  }

  my $deserialized = deserialize_cmd($msg);
  $msg = $deserialized;

  $msg =~ s/[\r\n]*$//;

#  printf("Rx Msg = %s\n",$msg);

  my ($status,$rest) = $msg =~ m/STATUS=([^\|]+)\|(.*)/;

  if (not ($status eq 'OK')) {
    job_stop_client($socket);
#    printf("Received not OK message from master (%s)",$msg);
    return $return_value;
  }

  $message = sprintf ("%s\n", $command);
#  printf("Message = %s",$message);
  job_send_data($socket,$message);

  $msg = <$socket>;

  if (not defined($msg)) {
    job_stop_client($socket);
#    printf("Timed out connecting to master to send command");
    return $return_value;
  }

  $deserialized = deserialize_cmd($msg);
  $msg = $deserialized;

  $msg =~ s/[\r\n]*$//;

#  printf("Rx Cmd Msg = %s\n",$msg);

  ($status,$rest) = $msg =~ m/STATUS=([^\|]+)\|(.*)/;

  if (not ($status eq 'OK')) {
    job_stop_client($socket);
#    printf("Received not OK message from master (%s)\n",$msg);
    return $return_value;
  }

  job_stop_client($socket);
#  printf("Statistics uploaded successfully\n");
  return 1;

}

sub GetMacAddress {

  my $ifconfigin= `/sbin/ifconfig eth0`;
  my @ifeth = split(/\n/, $ifconfigin);
  my ($junk,$macaddr) = split(/ HWaddr /, $ifeth[0]);

  # remove the ":" and spaces
  $macaddr =~ s/[\: ]//g;

#  printf "MAC=\"$macaddr\"\n";
  return $macaddr;
}


################################################################################
################################################################################

sub job_start_client {

  my $master_host_name = shift;
  my $master_host_port = shift;
  my $timeout = shift;
  my $silent = shift;


  $timeout = 60 if (not defined($timeout));
  $silent = 0 if (not defined($silent));


  my $arp_refresh = 1;
  if ($arp_refresh) {
    # Suspecting ARP cache timeout and subsequent connect failures as cause for
    # failures with "No Route to Host" error message
    # Hack it by massaging the ARP into clients ARP table with 3 pings.

#    debug(2,"JOB: Doing ping to host %s to refresh ARP cache...",$master_host_name);
    my $system_cmd =  sprintf ("ping -c 3 %s > %s 2>&1",$master_host_name,"/dev/null");
    my $execute = `$system_cmd`;
    if ($? == -1) {
#      debug(2,"ARP ping failed to exec: %s", $!);
    } elsif ($? >> 8) {
#      debug(2,"ARP ping exited with value %d\n", $? >> 8);
    }
  }

  my $socket = IO::Socket::INET->new(PeerAddr => $master_host_name,
                                     PeerPort => $master_host_port,
                                     Proto    => "tcp",
                                     Type     => SOCK_STREAM,
                                     Timeout  => $timeout);

  if (not defined($socket)) {
    if (not $silent) {
#      printf("Couldn't connect to Job Master = %s (TCP Port=%s) : %s\n",$master_host_name ,$master_host_port,$!);
    }
    return undef;
  } else {
    # Set timeout for socket
    $socket->timeout($timeout);
  }
  return $socket;
}

################################################################################
################################################################################

sub job_stop_client {

  my $client = shift;

  close($client) if (defined($client));
}

################################################################################
# The function serializes a complex datastructure into a safe and compact
# single line string.
################################################################################

sub serialize_cmd {
  my $cmd = shift;

  # replace all newlines, CR and % with CGI-style encoded sequences
  $cmd =~ s/([%\r\n])/sprintf("%%%02X", ord($1))/ge;

  return $cmd;
}

sub deserialize_cmd {
  my $serialized = shift;

  # convert back escapes to the original chars
  $serialized =~ s/%([0-9A-Fa-f]{2})/chr(hex($1))/ge;

  return $serialized;

}

################################################################################
################################################################################

sub job_send_data {

  my $socket = shift;
  my $message = shift;

  my $status = 0;
  if (defined($socket)) {
    my $msg = serialize_cmd($message);
    my $r = print $socket "$msg\n";
#    printf("Socket write error msg \'%s\' - %s",$msg,$!) if (!defined($r));
    $status = 1;
  }
  return $status;
}

################################################################################
################################################################################

sub job_receive_data {

  my $socket = shift;
  my $timeout = shift;

  $timeout = 1 if (not defined($timeout));

  my $msg = undef;
  if (defined($socket)) {

    # Set timeout for socket
    $socket->timeout($timeout);

    # Now wait timeout time for someone to send something
    my $client = $socket->accept();
    if (defined($client)) {
      $msg = <$client>;
      my $deserialized = deserialize_cmd($msg);
      return ($client, $deserialized);
    }
  }
  return undef;
}

